; #########################################################################
;
;                             qword2a
;
; Procedure for converting a 64-bit signed integer from memory to ASCII
; without any leading "0"
;
; Limitation: The st(6) and st(7) registers on the FPU will be trashed
;
; Return: EAX !=0 and null-terminated ASCII string in memory
;         EAX  =0 if FPU unusable and null string in memory
;
; Usage:  invoke qword2a,lpQW,lpBuf
;
;   where:  
;   lpQW  = pointer to the location of the qword in memory
;   lpBuf = pointer to a memory buffer for the null-terminated ASCII
;           string (a 21-byte buffer is sufficient for the largest of numbers
;           i.e. one sign, 19 digits maximum, and the terminating 0)
;
;                       by Raymond Filiatreault
;                            February 2007
;                            MASM  syntax
;
; #########################################################################
;
;      .386                   ; minimum processor needed for 32 bit
;      .model flat, stdcall   ; FLAT memory model & STDCALL calling
;      option casemap :none   ; set code to case sensitive
;
; #########################################################################

qword2a PROTO :DWORD, :DWORD

.code

qword2a proc USES esi edi lpQW:DWORD, lpBuf:DWORD

LOCAL oldcw       :WORD
LOCAL newcw       :WORD
LOCAL bcdstring   :TBYTE

      mov   edx,lpQW    ;EDX = address of qword
      mov   edi,lpBuf

;check first if qword = 0
      mov   eax,[edx]   ;get LOWDWORD
      or    eax,[edx+4] ;or it with HIGHDWORD
      jnz   @F
      mov   byte ptr[edi],"0"
      inc   edi
      jmp   eXit

   @@:
      fstcw oldcw       ;get control word
      fclex             ;clear exception flags
      ffree st(7)
      ffree st(6)       ;2 FPU registers required
      mov   ax,oldcw
      or    ah,0Fh      ;set precision to REAL10
                        ;and rounding control to truncating
      mov   newcw,ax
      fldcw newcw

      mov   eax,[edx+4] ;get the HIGHDWORD
      mov   ecx,[edx]   ;get the LOWDWORD
      test  eax,eax
      jns   @F
      neg   eax
      neg   ecx

;0DE0B6B3A7640000h = 10^18
   @@:
      cmp   eax,0DE0B6B3h
      pushf
      .if   ZERO?
            popf
            cmp   ecx,0A7640000h
      .else
            popf
      .endif

      .if   !CARRY?                       ;qword has 19 digits
            mov   eax,1000000000  ;10^9
            push  eax
            fild  dword ptr[esp]  ;load it on the FPU
            fmul  st,st       ;compute 10^18
            fild  qword ptr[edx] ;load the qword on the FPU
            fabs              ;make sure it is positive
            fdiv  st,st(1)    ;get the 19th digit if any
            frndint           ;keep only the integer
            fist  dword ptr[esp] ;store it on the stack
            fmul              ;that digit*10^18
            fild  qword ptr[edx] ;reload the qword on the FPU
            fabs
            fsubr             ;keeps only the lower 18 digits
      .else                               ;qword has less than 19 digits
            fild  qword ptr[edx] ;load the qword on the FPU
            xor   eax,eax
            fabs              ;make sure it is positive
            push  eax         ;19th digit = 0
      .endif
      fbstp bcdstring   ;transfer the 18 digits as a packed BCD
      fldcw oldcw       ;restore previous control word

      fstsw ax          ;get the FPU exception flags
      test  al,1        ;check for an invalid operation
      jz    @F
      pop   eax         ;clean the stack
      xor   eax,eax     ;return code for error
      jmp   eXit1

   @@:
      lea   esi,bcdstring
      mov   ecx,8
      mov   edx,[edx+4] ;get the HIGHDWORD of the qword
      test  edx,edx     ;to check the sign
      jns   @F
      mov   byte ptr[edi],"-"
      inc   edi
   @@:
      pop   eax         ;retrieve the 19th digit
      test  eax,eax
      jz    smaller_than_19
      add   al,30h      ;convert to ascii
      stosb             ;store first digit

;unpack the packed BCD and convert to ascii
bcd_to_a:
      xor   eax,eax
      mov   al,[esi+ecx]
      rol   eax,4
      ror   al,4
      add   ax,3030h
      xchg  al,ah
      stosw
      sub   ecx,1
      jnc   bcd_to_a
      jmp   eXit

smaller_than_19:
      mov   al,[esi+ecx]
      test  al,al
      jnz   @F          ;found first non-zero digits
      dec   ecx
      jmp   smaller_than_19
   @@:
      cmp   al,9
      ja    bcd_to_a    ;first non-zero byte has two digits
      add   al,30h      ;convert to ascii
      stosb             ;store first digit
      sub   ecx,1
      jnc   bcd_to_a    ;continue unpacking if more digits

eXit:
      or    al,1        ;no error code
eXit1:
      mov   byte ptr[edi],0  ;null string terminator
      ret

qword2a endp

;end
